from pysymphony.client import SymphonyClient
from pysymphony.workflows import Workflow
from pysymphony.runs import Run
from pysymphony.settings import settings

__all__ = ['SymphonyClient', 'Workflow', 'Run', 'settings']


import logging
import os
from typing import Dict, List

import requests
from dotenv import load_dotenv

from pysymphony.constants import ToolType
from pysymphony.tooling import LocalTool

load_dotenv()

class Settings:
    def __init__(
        self,
        session: requests.Session = None,
        local_tools: List[LocalTool] = [],
    ):
        assert os.getenv("SYMPHONY_API_KEY"), "SYMPHONY_API_KEY is not set"

        self.api_key = os.getenv("SYMPHONY_API_KEY")
        self.base_url = os.getenv("SYMPHONY_BASE_URL", "https://inference.composition-labs.com/inference")

        self.local_tools = local_tools

        self.session = session or requests.Session()
        self.session.headers.update({
            "Authorization": f"Bearer {self.api_key}"
        })

        self._tools = None

    @property
    def tools(self) -> List[str]:
        if not self._tools:
            res = self.get_tools()
            self._tools = []
            for tool in res:
                self._tools.append({
                    "id": tool["unique_id"],
                    "name": tool["name"],
                    "description": tool["description"],
                    "type": ToolType.SYMPHONY if tool["tool_type"] == "in_house" else ToolType.CLIENT
                })
            for tool in self.local_tools:
                self.add_client_tool(
                    unique_id=tool.id,
                    name=tool.name,
                    description=tool.description,
                    schema=tool.tool_schema,
                )
        return self._tools
    
    @tools.setter
    def tools(self, tools: List[str]):
        self._tools = tools

    def add_local_tool(
        self,
        fn: callable,
    ):
        """
        Add a local tool.
        """
        tool = LocalTool(fn=fn)
        self.local_tools.append(tool)
        self.add_client_tool(
            unique_id=tool.id,
            name=tool.name,
            description=tool.description,
            schema=tool.tool_schema,
        )

    def add_client_tool(
        self, 
        unique_id: str,
        name: str,
        description: str,
        schema: Dict,
    ):
        """
        Add a client tool.
        """
        tools = self.tools
        if unique_id in [tool["id"] for tool in tools]:
            return
        try:
            response = self.session.post(
                f"{self.base_url}/tools/add",
                json={
                    "unique_id": unique_id,
                    "name": name,
                    "description": description,
                    "schema": schema,
                },
            )
            if response.status_code == 200:
                logging.info(f"POST {self.base_url}/tools/add: SUCCESS")
                self.tools.append({
                    "id": unique_id,
                    "name": name,
                    "description": description,
                    "type": ToolType.CLIENT,
                })
                return
            else:
                logging.error(f"POST {self.base_url}/tools/add: ERROR {response.status_code} {response.text}")
                raise Exception(f"POST {self.base_url}/tools/add: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"POST {self.base_url}/tools/add: ERROR {e}")
            raise e
        
    def get_tools(self):
        """
        Get the tools.
        """
        try:
            response = self.session.get(f"{self.base_url}/tools/list")
            logging.info(f"GET {self.base_url}/tools/list: RUNNING")
            if response.status_code == 200:
                logging.info(f"GET {self.base_url}/tools/list: SUCCESS")
                return response.json()
            else:
                logging.error(f"GET {self.base_url}/tools/list: ERROR {response.status_code} {response.text}")
                raise Exception(f"GET {self.base_url}/tools/list: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"GET {self.base_url}/tools/list: ERROR {e}")
            raise e

settings = Settings()


import json
from inspect import signature
from typing import (Any, Awaitable, Callable, List, Optional, Tuple, Type,
                    Union, cast)

from pydantic import BaseModel, Field, model_validator, create_model
from pydantic.fields import FieldInfo


def create_schema_from_function(
    name: str,
    func: Union[Callable[..., Any], Callable[..., Awaitable[Any]]],
    additional_fields: Optional[
        List[Union[Tuple[str, Type, Any], Tuple[str, Type]]]
    ] = None,
) -> Type[BaseModel]:
    """Create schema from function."""
    fields = {}
    params = signature(func).parameters
    for param_name in params:
        param_type = params[param_name].annotation
        param_default = params[param_name].default

        if param_type is params[param_name].empty:
            param_type = Any

        if param_default is params[param_name].empty:
            # Required field
            fields[param_name] = (param_type, Field())
        elif isinstance(param_default, FieldInfo):
            # Field with pydantic.Field as default value
            fields[param_name] = (param_type, param_default)
        else:
            fields[param_name] = (param_type, Field(default=param_default))

    additional_fields = additional_fields or []
    for field_info in additional_fields:
        if len(field_info) == 3:
            field_info = cast(Tuple[str, Type, Any], field_info)
            field_name, field_type, field_default = field_info
            fields[field_name] = (field_type, Field(default=field_default))
        elif len(field_info) == 2:
            # Required field has no default value
            field_info = cast(Tuple[str, Type], field_info)
            field_name, field_type = field_info
            fields[field_name] = (field_type, Field())
        else:
            raise ValueError(
                f"Invalid additional field info: {field_info}. "
                "Must be a tuple of length 2 or 3."
            )
        
    return create_model(name, **fields).model_json_schema()

def get_desc_from_doc_str(fn: callable) -> str:
    """Get description from function."""
    docstring = str(fn.__doc__).strip()
    if docstring is None:
        raise ValueError("Function has no description, please add a description")
    return docstring


class LocalTool(BaseModel):
    fn: Callable
    id: Optional[str] = Field(default_factory=lambda: None)
    name: Optional[str] = Field(default_factory=lambda: None)
    description: Optional[str] = Field(default_factory=lambda: None)
    tool_schema: Optional[dict] = Field(default_factory=lambda: None)  # Renamed from 'schema'
    async_fn: Optional[Callable[..., Awaitable[Any]]] = Field(default_factory=lambda: None)

    @model_validator(mode='before')
    @classmethod
    def set_defaults(cls, values: dict) -> dict:
        fn = values.get('fn')
        if fn is None:
            return values

        if values.get('id') is None:
            values['id'] = fn.__name__

        if values.get('name') is None:
            values['name'] = values['id']

        if values.get('tool_schema') is None:  # Updated to match the new field name
            values['tool_schema'] = create_schema_from_function(values['name'], fn)

        if values.get('description') is None:
            values['description'] = get_desc_from_doc_str(fn)

        if values.get('async_fn') is None:
            values['async_fn'] = fn

        return values


import json
import logging

import requests
from pydantic import BaseModel

from pysymphony.constants import Status
from pysymphony.settings import settings


class ToolCallInfo(BaseModel):
    name: str
    input: str
    output: str | None = None

class ToolCalls(BaseModel):
    calls: dict[str, ToolCallInfo]

class Run:
    def __init__(
        self, 
        id: str, 
        workflow_id: str, 
        input: str, 
        session: requests.Session = settings.session,
        base_url: str = settings.base_url,
        status: Status = Status.IDLE,
    ):
        self.id = id
        self.workflow_id = workflow_id
        self.input = input
        self.base_url = base_url
        self.session = session or requests.Session()

        self.status = status

        self.response = None
        self.ios = None
        self.tool_calls = None

    def __str__(self):
        return f"Run(id={self.id}, workflow_id={self.workflow_id}, input={self.input})"
    
    def update_status(self, status: Status):
        self.status = status

    def run_step(self):
        """
        Run a step.
        """
        try:
            self.update_status(Status.RUNNING)
            json_input = {
                "run_id": self.id
            }
            if self.tool_calls is not None:
                json_input["tool_calls"] = {
                    key: str(t.output) for key, t in self.tool_calls.calls.items()
                }
                self.tool_calls = None
            response = self.session.post(
                f"{self.base_url}/runner/runstep",
                json=json_input
            )
            if response.status_code == 200:
                logging.info(f"POST {self.base_url}/runner/runstep: SUCCESS")
                res = response.json()

                if res["tool_calls"]:
                    self.tool_calls = ToolCalls(calls=res["tool_calls"])
                    self.update_status(Status.WAITING_FOR_CLIENT_TOOL)
                    self.handle_tool_calls()
                    self.update_status(Status.RUNNING)
                if res["status"] == "completed":
                    self.update_status(Status.SUCCESS)
                self.ios = res["ios"]
            else:
                logging.error(f"POST {self.base_url}/runner/runstep: ERROR {response.status_code} {response.text}")
                raise Exception(f"POST {self.base_url}/runner/runstep: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"POST {self.base_url}/runner/runstep: ERROR {e}")
            self.update_status(Status.ERROR)
            raise e

    async def arun_step(self, input: str):
        """
        Run a step.
        """
        raise NotImplementedError("`arun_step` is not implemented yet")
    
    def run(
        self,
    ):
        """
        Run a workflow.
        """
        while self.status != Status.SUCCESS and self.status != Status.ERROR:
            self.run_step()
        return self.status

    async def arun(self, workflow_id: str):
        """
        Run a workflow.
        """
        while self.status != Status.SUCCESS and self.status != Status.ERROR:
            await self.arun_step()
        return self.status

    def handle_tool_call(self, tool_call: ToolCallInfo):
        """
        Handle a tool call.
        """
        local_tool = [tool for tool in settings.local_tools if tool_call.name == tool.id]
        if len(local_tool) == 0:
            raise Exception(f"Tool {tool_call.name} not found in local tools. Please register the tool first.")
        local_tool = local_tool[0]
        res = local_tool.fn(**json.loads(tool_call.input))
        tool_call.output = res
    
    def handle_tool_calls(self):
        """
        Handle tool calls.
        """
        for tool_call in self.tool_calls.calls.values():
            if tool_call.output is None:
                self.handle_tool_call(tool_call)

    async def ahandle_tool_calls(self):
        """
        Handle tool calls.
        """
        raise NotImplementedError("`ahandle_tool_calls` is not implemented yet")

    async def ahandle_tool_call(self, tool_call: ToolCallInfo):
        """
        Handle a tool call.
        """
        raise NotImplementedError("`ahandle_tool_call` is not implemented yet")



import logging
from typing import Any, Dict, List

import requests

from pysymphony.runs import Run
from pysymphony.settings import settings
from pysymphony.utils import get_mermaid_image_url


def resolve_parameters(parameters: Dict[str, Any]):
    """
    Resolve the parameters.
    """
    for key in parameters.keys():
        if key not in ["llmmodel", "max_steps", "optimize", "max_optimization_steps"]:
            raise ValueError(f"Parameter {key} is not supported")
    return parameters

class Workflow:
    def __init__(
        self, 
        task_description: str = None,
        id: str = None, 
        nodes: List[Dict[str, Any]] = [],
        runs: List[Run] = [],
        parameters: Dict[str, Any] = {},
        session: requests.Session = settings.session,
        tools: List[str] = [],
        base_url: str = settings.base_url,
    ):
        self.id = id
        
        self.task_description = task_description
        self.nodes = nodes
        self.runs = runs
        self.parameters = parameters
        self.session = session or requests.Session()
        self.tools = tools
        self.base_url = base_url

    def __str__(self):
        return f"Workflow(id={self.id}, task_description={self.task_description})"
    
    def run_workflow(self, input: str):
        """
        Run the workflow.
        """
        raise NotImplementedError("`run_workflow` is not implemented yet, use `init_run` and `run_step` instead")
        # if not self.id:
        #     raise ValueError("Workflow is not generated yet")
        # self.runs.append(Run(
        #     id=self.id,
        #     workflow_id=self.id,
        #     input=input,
        #     session=self.session,
        #     base_url=self.base_url,
        # ))
        # _ = self.runs[-1].run()
        # return self.runs[-1]
    
    async def arun_workflow(self, input: str):
        """
        Run the workflow.
        """
        raise NotImplementedError("`run_workflow` is not implemented yet, use `init_run` and `run_step` instead")
    
    def init_run(self, input: str):
        """
        Initialize a run.
        """
        try:
            response = self.session.post(
                f"{self.base_url}/runner/initrun",
                json={
                    "workflow_id": self.id,
                    "text_input": input,
                },
            )
            if response.status_code == 200:
                logging.info(f"POST {self.base_url}/runner/initrun: SUCCESS")
                res = response.json()
                return Run(
                    id=res["run_id"],
                    workflow_id=self.id,
                    input=input,
                    session=self.session,
                    base_url=self.base_url,
                )
            else:
                logging.error(f"POST {self.base_url}/runner/initrun: ERROR {response.status_code} {response.text}")
                raise Exception(f"POST {self.base_url}/runner/initrun: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"POST {self.base_url}/runner/initrun: ERROR {e}")
            raise e
    
    def run_step(self, input: str):
        """
        Run a step.
        """
        pass

    def generate_workflow(
        self
    ):
        """
        Generate a workflow.
        """
        try:
            response = self.session.post(
                f"{self.base_url}/planner/generateworkflow",
                json={
                    "belongs_to": "",
                    "task_description": self.task_description,
                    "tools": self.tools,
                    "parameters": resolve_parameters(self.parameters),
                },
            )
            if response.status_code == 200:
                logging.info(f"POST {self.base_url}/planner/generateworkflow: SUCCESS")
                res = response.json()
                self.id = res["id"]
                self.nodes = res["nodes"]
                self.parameters = res["parameters"]
            else:
                logging.error(f"POST {self.base_url}/planner/generateworkflow: ERROR {response.status_code} {response.text}")
                raise Exception(f"POST {self.base_url}/planner/generateworkflow: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"POST {self.base_url}/planner/generateworkflow: ERROR {e}")
            raise e

    async def agenerate_workflow(
        self
    ):
        """
        Generate a workflow.
        """
        raise NotImplementedError("`generate_workflow` is not implemented yet")
    
    def get_mermaid_image_url(self):
        """
        Get the mermaid image url.
        """
        return get_mermaid_image_url(self.get_mermaid_graph())
    
    def get_mermaid_graph(self):
        """
        Get the mermaid graph.
        """
        mm_str = f"graph TD\nclassDef bg-rounded fill:#f9f,stroke:#333,stroke-width:2px,rx:10px,ry:10px;\n"
        mm_str += f"INPUT\n"
        mm_str += "OUTPUT[Output]\n"
        for node in self.nodes:
            content = node["job"]
            if len(content) > 100:
                content = content[:100] + "..."
            # for tool in node["tools"]:
            #     content += f"\n<br>**{tool['name']}**"
            mm_str += f"Step_{str(node['step_index'])}[{content}]\n"
        for node in self.nodes:
            if len(node["tools"]) > 0:
                tool_content = ""
                for tool in node["tools"]:
                    tool_content += f"\n<br><span class='bg-rounded'>**{tool['name']}**</span>"
                mm_str += f"Step_{str(node['step_index'])} -.-> tools_{str(node['step_index'])}[{tool_content}]\n"
                mm_str += f"tools_{str(node['step_index'])} --> Step_{str(node['step_index'])}\n"
            dependencies = node["dependencies"]
            for dependency in dependencies:
                mm_str += f"Step_{str(dependency)} --> Step_{str(node['step_index'])}\n"
            if len(dependencies) == 0:
                mm_str += f"INPUT --> Step_{str(node['step_index'])}\n"
        for node in self.nodes:
            if node["step_index"] == len(self.nodes):
                mm_str += f"Step_{str(node['step_index'])} --> OUTPUT\n"
        return mm_str



from enum import Enum

class Status(Enum):
    IDLE = "idle"
    RUNNING = "running"
    SUCCESS = "success"
    ERROR = "error"
    WAITING_FOR_CLIENT_TOOL = "waiting_for_client_tool"

class ToolType(Enum):
    SYMPHONY = "symphony"
    CLIENT = "client"
    API = "api"


import logging
from typing import Any, Dict, List

import requests

from pysymphony.tooling import LocalTool
from pysymphony.workflows import Workflow
from pysymphony.settings import settings


class SymphonyClient:
    """
    SymphonyClient is a client for the Symphony API.
    """
    def __init__(
        self,
        api_key: str = settings.api_key,
        base_url: str = settings.base_url,
        session: requests.Session = settings.session,
        local_tools: List[LocalTool] = [],
    ):
        """
        Initialize the SymphonyClient.
        """
        if not settings.api_key:
            raise ValueError("API key is required")

        self.api_key = api_key
        self.base_url = base_url
        self.session = session or requests.Session()

        self.session.headers.update({
            "Authorization": f"Bearer {self.api_key}"
        })

        self.local_tools = local_tools
        self.tools = settings.tools
        
        self._workflows = []

    @property
    def workflows(self) -> List[str]:
        if not self._workflows:
            res = self.get_workflows()
            self._workflows = []
            for workflow in res:
                self._workflows.append(Workflow(
                    id=workflow["id"],
                    task_description=workflow["task_description"],
                    nodes=workflow["nodes"],
                    parameters=workflow["parameters"],
                    session=self.session,
                    base_url=self.base_url,
                ))
        return self._workflows
    
    @workflows.setter
    def workflows(self, workflows: List[str]):
        self._workflows = workflows

    async def aget_tools(self):
        """
        Get the tools.
        """
        pass

    def get_workflows(self):
        """
        Get the workflows.
        """
        try:
            response = self.session.get(f"{self.base_url}/planner/getworkflows")
            logging.info(f"GET {self.base_url}/planner/getworkflows: RUNNING")
            if response.status_code == 200:
                logging.info(f"GET {self.base_url}/planner/getworkflows: SUCCESS")
                return response.json()
            else:
                logging.error(f"GET {self.base_url}/planner/getworkflows: ERROR {response.status_code} {response.text}")
                raise Exception(f"GET {self.base_url}/planner/getworkflows: ERROR {response.status_code} {response.text}")
        except Exception as e:
            logging.error(f"GET {self.base_url}/planner/getworkflows: ERROR {e}")
            raise e

    async def aget_workflows(self):
        """
        Get the workflows.
        """
        pass

    def generate_workflow(
        self,
        task_description: str,
        parameters: Dict[str, Any],
        tools: List[str] = [],
    ):
        """
        Generate a workflow.
        """
        workflow = Workflow(task_description=task_description, parameters=parameters, tools=tools, session=self.session, base_url=self.base_url)
        workflow.generate_workflow()
        self.workflows.append(workflow)
        return workflow
    
    async def agenerate_workflow(
        self,
        task_description: str,
        parameters: Dict[str, Any],
        tools: List[str] = [],
    ):
        """
        Generate a workflow.
        """
        pass


import base64


def get_mermaid_image_url(graph: str) -> str:
    graphbytes = graph.encode("ascii")
    base64_bytes = base64.b64encode(graphbytes)
    base64_string = base64_bytes.decode("ascii")
    return "https://mermaid.ink/img/" + base64_string


from setuptools import setup, find_packages

setup(
    name='pysymphony',
    version='0.2.3',
    description='A Python client for Symphony',
    author='Chinmay Shrivastava',
    author_email='cshrivastava99@gmail.com',
    url='https://github.com/compositionlabs/pysymphony',
    packages=find_packages(),
    install_requires=[
        'certifi==2024.8.30',
        'charset-normalizer==3.3.2',
        'idna==3.10',
        'requests==2.32.3',
        'urllib3==2.2.3',
    ],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
)
